use itertools::Either;
use std::collections::HashMap;
use std::fmt::{Display, Write};
use std::fs::File;
use std::io::{BufRead, BufReader, Lines};
use std::time::Instant;

// BOILERPLATE
type InputIter = Lines<BufReader<File>>;

fn get_input() -> InputIter {
    let f = File::open("input").unwrap();
    let br = BufReader::new(f);
    br.lines()
}

fn main() {
    let start = Instant::now();
    let ans1 = problem1(get_input());
    let duration = start.elapsed();
    println!("Problem 1 solution: {} [{}s]", ans1, duration.as_secs_f64());

    let start = Instant::now();
    let ans2 = problem2(get_input());
    let duration = start.elapsed();
    println!("Problem 2 solution: {} [{}s]", ans2, duration.as_secs_f64());
}

// PARSE

#[derive(Debug)]
struct Platform {
    rows: Vec<Vec<char>>,
    width: usize,
    height: usize,
}

impl<T: BufRead> From<Lines<T>> for Platform {
    fn from(lines: Lines<T>) -> Self {
        let rows: Vec<Vec<char>> = lines.map(|line| line.unwrap().chars().collect()).collect();
        let width = rows[0].len();
        let height = rows.len();
        Self {
            rows,
            width,
            height,
        }
    }
}

#[derive(Debug, PartialEq)]
enum Axis {
    Rows,
    Columns,
}

struct North;
struct South;
struct East;
struct West;

trait Direction {
    const REVERSED: bool;
    const AXIS: Axis;
}

impl Direction for North {
    const REVERSED: bool = false;
    const AXIS: Axis = Axis::Columns;
}
impl Direction for South {
    const REVERSED: bool = true;
    const AXIS: Axis = Axis::Columns;
}
impl Direction for East {
    const REVERSED: bool = true;
    const AXIS: Axis = Axis::Rows;
}
impl Direction for West {
    const REVERSED: bool = false;
    const AXIS: Axis = Axis::Rows;
}

impl<'a> Platform {
    fn col(&self, col: usize) -> Vec<char> {
        self.rows.iter().map(|row| row[col]).collect()
    }
    fn replace_col(&mut self, idx: usize, col: Vec<char>) {
        for row_idx in 0..col.len() {
            self.rows[row_idx][idx] = col[row_idx];
        }
    }
    fn roll<T: Direction>(&mut self) {
        match T::AXIS {
            Axis::Columns => {
                for idx in 0..self.width {
                    let new = self.roll_one::<T>(idx);
                    self.replace_col(idx, new);
                }
            }
            Axis::Rows => {
                for idx in 0..self.height {
                    let new = self.roll_one::<T>(idx);
                    self.rows[idx].copy_from_slice(new.as_slice());
                }
            }
        }
    }
    fn roll_rock_at<T: Direction>(old: &mut Vec<char>, inner_idx: usize) {
        // Find the first # rock or the edge of the map, we can't look beyond this for a resting position
        if T::REVERSED {
            // The furthest possible resting position is the first # below us, or the edge of the map
            let upper_limit = (inner_idx..old.len() - 1)
                .find(|c| old[*c] == '#')
                .unwrap_or(old.len());

            // We will roll to the position of the furthest '.' before the upper limit
            if let Some(empty_pos) = (inner_idx..upper_limit).rev().find(|i| old[*i] == '.') {
                old.swap(inner_idx, empty_pos);
            }
        } else {
            // the furthest possible resting position is the first # above us, or the edge
            let lower_limit = (0..inner_idx).rev().find(|c| old[*c] == '#').unwrap_or(0);

            // We will roll to the position of the furthest '.' before the lower limit
            if let Some(empty_pos) = (lower_limit..inner_idx).find(|i| old[*i] == '.') {
                old.swap(inner_idx, empty_pos);
            }
        }
    }

    fn roll_one<T: Direction>(&mut self, idx: usize) -> Vec<char> {
        let mut old = match T::AXIS {
            Axis::Columns => self.col(idx),
            Axis::Rows => self.rows[idx].clone(),
        };
        let idx_iter: Either<_, _> = match T::REVERSED {
            true => Either::Left((0..old.len()).rev()),
            false => Either::Right(0..old.len()),
        };
        let end_idx = if T::REVERSED { old.len() } else { 0 };
        for inner_idx in idx_iter {
            match old[inner_idx] {
                '.' | '#' => continue,
                'O' if inner_idx == end_idx => continue,
                'O' => Self::roll_rock_at::<T>(&mut old, inner_idx),
                _ => panic!("invalid character"),
            }
        }
        old
    }
    fn score<T: Direction>(&self) -> u64 {
        match T::AXIS {
            Axis::Columns => self.score_columns::<T>(),
            Axis::Rows => self.score_rows::<T>(),
        }
    }
    fn row_or_column_score<T: Direction>(&self, idx: usize) -> usize {
        if T::REVERSED {
            idx + 1
        } else {
            match T::AXIS {
                Axis::Rows => self.width - idx,
                Axis::Columns => self.height - idx,
            }
        }
    }
    fn score_columns<T: Direction>(&self) -> u64 {
        (0..self.height)
            .map(|row| {
                self.rows[row].iter().filter(|c| **c == 'O').count()
                    * self.row_or_column_score::<T>(row)
            })
            .sum::<usize>() as u64
    }
    fn roll_cycle(&mut self) {
        self.roll::<North>();
        self.roll::<West>();
        self.roll::<South>();
        self.roll::<East>();
    }

    // find the first loop, return the iteration count when we first saw it and when we saw it again
    fn find_loop(&mut self) -> (usize, usize) {
        let mut first_seen: HashMap<Vec<Vec<char>>, usize> = HashMap::new();
        first_seen.insert(self.rows.clone(), 0);
        let mut i = 0;
        loop {
            self.roll_cycle();
            i += 1;
            if let Some(first_idx) = first_seen.insert(self.rows.clone(), i) {
                return (first_idx, i);
            }
        }
    }
    fn score_rows<T: Direction>(&self) -> u64 {
        unimplemented!()
    }
}

impl Display for Platform {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for row in &self.rows {
            for c in row {
                f.write_char(*c)?;
            }
            writeln!(f)?;
        }
        writeln!(f)
    }
}

// PROBLEM 1 solution

fn problem1<T: BufRead>(input: Lines<T>) -> u64 {
    let mut p = Platform::from(input);
    p.roll::<North>();
    p.score::<North>()
}

// PROBLEM 2 solution
fn problem2<T: BufRead>(input: Lines<T>) -> u64 {
    const ITERATIONS: usize = 1000000000;
    let mut p = Platform::from(input);
    let first_loop = p.find_loop();
    let loop_length = first_loop.1 - first_loop.0;
    let cycles_to_skip = ((ITERATIONS - first_loop.0) / loop_length) * loop_length;
    let iterations_remaining = ITERATIONS - first_loop.0 - cycles_to_skip;

    for _ in 0..iterations_remaining {
        p.roll_cycle();
    }

    p.score::<North>()
}

#[cfg(test)]
mod tests {
    use crate::*;
    use std::io::Cursor;

    const EXAMPLE: &str = &"O....#....
O.OO#....#
.....##...
OO.#O....O
.O.....O#.
O.#..O.#.#
..O..#O..O
.......O..
#....###..
#OO..#....";

    #[test]
    fn problem1_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem1(c.lines()), 136);
    }

    #[test]
    fn problem2_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem2(c.lines()), 64);
    }
}
